"""
Assemble a topology diagram.
"""
from ghpythonlib.componentbase import executingcomponent as component

from compas_cem.diagrams import TopologyDiagram


class AssembleTopologyDiagramComponent(component):
    def RunScript(self, trail_edges, deviation_edges, loads, supports, add_auxiliary_trails):

        add_auxiliary_trails = add_auxiliary_trails or False
        topology = TopologyDiagram()

        if trail_edges:
            for trail_edge in trail_edges:
                if trail_edge:
                    topology.add_edge(trail_edge)

        if deviation_edges:
            for deviation_edge in deviation_edges:
                if deviation_edge:
                    topology.add_edge(deviation_edge)

        if loads:
            for load in loads:
                if load:
                    topology.add_load(load)

        if supports:
            for support in supports:
                if support:
                    topology.add_support(support)

        if trail_edges and supports:
            topology.build_trails(add_auxiliary_trails)

        return topology



from math import copysign
from math import pi

import matplotlib.pyplot as plt

from compas.geometry import add_vectors
from compas.geometry import rotate_points_xy
from compas.geometry import length_vector
from compas.geometry import scale_vector
from compas.geometry import translate_points
from compas.geometry import normalize_vector

from compas_plotters import NetworkPlotter

from compas_cem import COLORS


__all__ = ["TopologyPlotter"]


class TopologyPlotter(NetworkPlotter):
    """
    A plotter tailored to draw topology matters.

    Parameters
    ----------
    topology_diagram : :class:`compas_cem.diagrams.TopologyDiagram`
        The topology diagram to plot.
    """
    def __init__(self, topology_diagram, *args, **kwargs):
        super(TopologyPlotter, self).__init__(topology_diagram, *args, **kwargs)

        self._node_colors = {"support": COLORS["node_support"],
                             "_origin": COLORS["node_origin"],
                             "default": COLORS["node"]}

        self._edge_state_colors = {-1.0: COLORS["compression"],
                                   1.0: COLORS["tension"],
                                   0.0: COLORS["edge"],
                                   "auxiliary_trail": COLORS["auxiliary_trail"]}

        self._edge_linestyles = {"trail": "-",  # solid
                                 "deviation": "--"}  # dashed

        self._topology = self.datastructure
        self._float_precision = "3f"

    @property
    def topology(self):
        """
        The topology diagram to plot.

        Returns
        -------
        topology : :class:`compas_cem.diagrams.TopologyDiagram`
            A topology diagram.
        """
        return self._topology

    @property
    def node_colors(self):
        """
        The colors to draw the nodes of a topology diagram.

        Returns
        -------
        node_colors : ``dict`` of ``color``
            A dictionary that maps node type to RGB color.
        """
        return self._node_colors

    @property
    def edge_state_colors(self):
        """
        The colors to draw edges based on the forces acting on them.

        Returns
        -------
        edge_state_colors : ``dict`` of ``color``
            A dictionary that maps edge force to RGB color.
        """
        return self._edge_state_colors

    @property
    def edge_linestyles(self):
        """
        The linestyles used to draw edges based on the their topological type.

        Returns
        -------
        edge_linestyles: ``dict`` of ``str``
            A dictionary that maps edge type to linestyle.
        """
        return self._edge_linestyles

    @property
    def float_precision(self):
        """
        The default decimal precision to render float values as text.

        Returns
        -------
        precision: ``str``
            The float precision value. Defaults to ``3f``.
        """
        return self._float_precision

    def draw_nodes(self, *args, **kwargs):
        """
        Draws the nodes of a ``TopologyDiagram``.

        Parameters
        ----------
        keys : ``list`` of ``int``
            The keys of the nodes to plot.
        radius : ``float``, ``dict`` of ``float``
            The radius of the nodes.
        text : ``str``, ``dict`` of ``str``
            A dictionary of strings to render on the nodes.
        facecolor : ``color``, ``dict`` of ``color``
            Color for the node circle fill in (r, g, b) format.
        edgecolor : ``color``, ``dict`` of ``color``
            Color for the node circle edge in (r, g, b) format.
        edgewidth : ``float``, ``dict`` of ``float``
            Width for the node circle edge.
        textcolor : ``color``, ``dict`` of ``color``
            Color for the text to be displayed on the nodes.
        fontsize : ``int``, ``dict`` of ``int``
            Font size for the text to be displayed on the nodes.

        Returns
        -------
        collection : ``matplotlib.collection``
            A matplotlib point collection object.

        Notes
        -----
        When the parameters are passed as single value, this will be applied to
        all the nodes or edges in the ``TopologyDiagram``.
        If instead, a dictionary that maps ``{node_key: attribute}`` is given,
        specific values can be assigned individually.
        """
        cmap = self.node_colors
        ds = self.datastructure
        cmap["d"] = (255, 255, 255)  # default color
        nc = {n: cmap[ds.node_attribute(n, "type") or "d"] for n in ds.nodes()}

        super(TopologyPlotter, self).draw_nodes(facecolor=nc, *args, **kwargs)

    def draw_edges(self, *args, **kwargs):
        """
        Draws the edges of a ``TopologyDiagram``.

        Parameters
        ----------
        keys : ``list`` of ``tuple``
            The keys of the edges to plot.
            Defaults to ``None``.
        width : ``float``, ``dict`` of ``float``
            Width of the edges.
        color : ``color``, ``dict`` of ``color``
            Color of the edges in (r, g, b) format.
        text : ``str``, ``dict`` of ``str``
            A dictionary of strings to render on the nodes.
        textcolor : ``color``, ``dict`` of ``color``
            Color for the text to be displayed on the nodes.
        fontsize : ``int``, ``dict`` of ``int``
            Font size for the text to be displayed on the nodes.

        Returns
        -------
        collection : ``matplotlib.collection``
            A matplotlib point collection object.

        Notes
        -----
        When the parameters are passed as single value, this will be applied to
        all the nodes or edges in the ``TopologyDiagram``.
        If instead, a dictionary that maps ``{edge_key: attribute}`` is given,
        specific values can be assigned individually.
        """
        ds = self.datastructure
        cmap = self.edge_state_colors

        keys = list(ds.edges())
        aux_keys = [tuple(aux_edge) for aux_edge in ds.auxiliary_trails()]

        ec = {}
        for edge in keys:
            if edge in aux_keys:
                ec[edge] = cmap["auxiliary_trail"]
                continue

            if ds.is_trail_edge(edge):
                attr_name = "length"
            else:
                attr_name = "force"
            ec[edge] = cmap[copysign(1.0, ds.edge_attribute(edge, attr_name))]

        edges = super(TopologyPlotter, self).draw_edges(color=ec, *args, **kwargs)

        lsmap = self.edge_linestyles
        els = [lsmap[ds.edge_attribute(e, "type")] for e in keys]
        edges.set_linestyle(els)

        return ec

    def draw_loads(self, keys=None, radius=0.1, width=0.5, draw_arrows=False, **kwargs):
        """
        Draws the node loads as crosses inscribed in a circle.

        Parameters
        ----------
        keys : ``list``
            The list of node keys where to draw forces.
            If nodes is ``None``, all nodes in will be considered.
            Defaults to ``None``.
        radius : ``float``, ``dict`` of ``float``
            The radius of the loads. Defaults to ``0.1``.
        width : ``float``
            The arrows uniform display width. Defaults to ``0.5``.
        draw_arrows : ``bool``
            A flag to draw the loads as arrows. Defaults to ``False``.
        kwargs : ``dict``
            The named attributes to display the the arrows. Defaults to ``None``.
        """
        topology = self.datastructure
        flips = (-1, 1)
        angle = pi / 4.0

        if not keys:
            keys = list(topology.nodes())

        loads = []
        for node in keys:
            if not topology.is_node_loaded(node):
                continue

            xyz = topology.node_coordinates(node)
            line = [add_vectors(xyz, [radius * f, 0.0, 0.0]) for f in flips]

            for f in flips:
                start, end = rotate_points_xy(line, angle=f*angle, origin=xyz)

                load = {}
                load['start'] = start
                load['end'] = end
                load['color'] = (0, 0, 0)  # black
                load['width'] = width

                loads.append(load)

        lines = self.draw_lines(loads)
        lines.set_zorder(4000)

        if draw_arrows:
            self._draw_load_arrows(keys=keys, width=width, **kwargs)

        return lines

    def _draw_load_arrows(self, keys=None, scale=1.0, width=1.0, gap=0.05, tol=1e-3):
        """
        Draws the node loads in a ``TopologyDiagram`` as scaled arrows.

        Parameters
        ----------
        keys : ``list``
            The list of node keys where to draw forces.
            If nodes is ``None``, all nodes in will be considered.
            Defaults to ``None``.
        scale : ``float``
            The scale of the load arrows. Defaults to ``1.0``.
        width : ``float``
            The arrows uniform display width. Defaults to ``1.0``.
        gap : ``float``
            The offset between the node and the load. Defaults to ``0.2``.
        tol : ``float``
            The minimum force magnitude to draw. Defaults to ``1e-3``.
        """
        keys = keys or list(self.datastructure.nodes())
        attrs = ["qx", "qy", "qz"]
        color = COLORS["load"]
        shift = {key: False for key in keys}
        self._draw_forces(keys, attrs, scale, color, width, shift, gap, tol)

    def draw_segments(self, segments, color=(50, 50, 50), width=0.5, ls="--"):
        """
        Draws additional line segments on a ``TopologyDiagram``.

        Parameters
        ----------
        segments : ``list``
            The line segments as tuples of xyz coordinates.
        color : ``tuple``
            The lines uniform color in rgb. Defaults to gray, ``(40, 40, 40)``.
        width : ``float``
            The lines' uniform display width. Defaults to ``0.5``.
        """
        lines = []
        for segment in segments:
            line = {}
            start, end = segment

            line["start"] = start
            line["end"] = end
            line["color"] = color
            line["width"] = width

            lines.append(line)

        lines = super(TopologyPlotter, self).draw_lines(lines)
        lines.set_linestyle(ls)

        return lines

    def save(self, filepath, tight=True, autoscale=True, bbox_inches="tight", pad_inches=0.0, **kwargs):
        """
        Saves the plot to a file.

        Parameters
        ----------
        filepath : str
            Full path of the file.
        """
        if autoscale:
            self.axes.autoscale(tight=False)

        if tight:
            plt.tight_layout()
            kwargs_tight = {"bbox_inches": bbox_inches, "pad_inches": pad_inches}
            kwargs.update(kwargs_tight)

        plt.savefig(filepath, **kwargs)

    def _draw_forces(self, keys, attrs, scale, color, width, shift, gap, tol):
        """
        Base method to draw forces (residuals or loads) as scaled arrows.

        Parameters
        ----------
        keys : ``list``
            The list of node keys where to draw forces.
        attrs : ``list``
            The attribute names of the force vector to draw.
        scale : ``float``
            The forces scale factor.
        color : ``tuple``
            The forces' color in rgb.
        width : ``float``
            The forces  display width.
        shift : ``bool``
            A flat to shift an arrow one length along its own axis.
        gap : ``float``
            The offset between the node and the force.
        tol : ``float``
            The minimum force magnitude to draw.
        """
        ds = self.datastructure
        arrows = []

        for node in keys:
            q_vec = ds.node_attributes(node, attrs)
            q_vec_scaled = scale_vector(q_vec, scale)
            q_vec_norm = normalize_vector(q_vec)
            q_len = length_vector(q_vec)

            if q_len < tol:
                continue

            arrow = {}
            start = ds.node_xyz(node)
            end = add_vectors(start, q_vec_scaled)

            # shift
            gap_arrow = gap
            if shift[node]:
                gap_arrow = (gap + length_vector(q_vec_scaled)) * -1

            gap_vector = scale_vector(q_vec_norm, gap_arrow)
            start, end = translate_points([start, end], gap_vector)

            # create gap
            arrow["start"] = start
            arrow["end"] = end
            arrow["color"] = color
            arrow["width"] = width

            arrows.append(arrow)

        super(TopologyPlotter, self).draw_arrows(arrows)


if __name__ == "__main__":
    pass



"""
compas_cem.diagrams
****************************

.. currentmodule:: compas_cem.diagrams


Diagrams
========

.. autosummary::
    :toctree: generated/
    :nosignatures:

    TopologyDiagram
    FormDiagram
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


# from .<module> import *
from .mixins import *  # noqa F403
from .diagram import *  # noqa F403
from .topology import *  # noqa F403
from .form import *  # noqa F403


__all__ = [name for name in dir() if not name.startswith('_')]



"""
Generate a form diagram in static equilibrium.
"""
from ghpythonlib.componentbase import executingcomponent as component

from compas_cem import PROXY_PORT

from compas.rpc import Proxy


class ConstrainedFormFindingComponent(component):
    def RunScript(self, topology, constraints, parameters, algorithm, iters_max, eps, tmax, eta):
        algorithm = algorithm or "LD_SLSQP"
        iters_max = iters_max or 100
        eps = eps or 1e-6
        tmax = tmax or 100
        eta = eta or 1e-6

        if topology and constraints and parameters:
            topology = topology.copy()

            # clean constraints and parameters from None
            constraints = [c for c in constraints if c is not None]
            parameters = [p for p in parameters if p is not None]

            with Proxy("compas_cem.optimization", port=PROXY_PORT) as opt:
                solution = opt.solve_nlopt_proxy(topology,
                                                 constraints,
                                                 parameters,
                                                 algorithm,
                                                 iters_max,
                                                 eps,
                                                 tmax,
                                                 eta)

            form, objective, grad_norm, iters, time, status = solution
            return form, objective, grad_norm, iters, time, status



"""
Disassemble a topology diagram into its constituent parts.
"""
from ghpythonlib.componentbase import executingcomponent as component


class TopologyDisassemblyComponent(component):
    def RunScript(self, topology):
        if topology:
            trail_keys = list(topology.trails())
            auxiliary_trail_keys = list(topology.auxiliary_trails())

            edge_keys = list(topology.edges())
            trail_edge_keys = list(topology.trail_edges())
            deviation_edge_keys = list(topology.deviation_edges())

            node_keys = list(topology.nodes())
            origin_node_keys = list(topology.origin_nodes())
            support_node_keys = list(topology.support_nodes())

            return node_keys, origin_node_keys, support_node_keys, edge_keys, trail_edge_keys, deviation_edge_keys, trail_keys, auxiliary_trail_keys



"""
Make the force on a trail edge to reach a goal force value.
"""
from ghpythonlib.componentbase import executingcomponent as component

from compas_cem.optimization import TrailEdgeForceConstraint


class TrailEdgeForceConstraintComponent(component):
    def RunScript(self, edge_key, force, weight):
        weight = weight or 1.0
        if edge_key and force is not None:
            return TrailEdgeForceConstraint(edge_key, force, weight)



"""
Disassemble a form diagram into its constituent parts.
"""
from ghpythonlib.componentbase import executingcomponent as component


class FormDisassemblyComponent(component):
    def RunScript(self, form):
        if form:
            edge_keys = list(form.edges())
            node_keys = list(form.nodes())
            support_node_keys = list(form.support_nodes())

            return node_keys, support_node_keys, edge_keys



"""
Draw a form diagram.
"""
from ghpythonlib.componentbase import executingcomponent as component

from compas_cem.ghpython import FormArtist


class FormArtistComponent(component):
    def RunScript(self, form, node_keys, edge_keys, force_min, force_scale):
        node_keys = node_keys or None
        edge_keys = edge_keys or None
        force_min = force_min or 1e-3
        force_scale = force_scale or 1
        if form:
            artist = FormArtist(form)

            nodes = artist.draw_nodes(node_keys)
            edges = artist.draw_edges(edge_keys)
            support_nodes = artist.draw_nodes_support(node_keys)

            loads = artist.draw_loads(node_keys, min_load=force_min, scale=force_scale)
            reactions = artist.draw_reactions(node_keys, min_force=force_min, scale=force_scale)

            return nodes, support_nodes, edges, loads, reactions



"""
compas_cem.diagrams
****************************

.. currentmodule:: compas_cem.diagrams


Classes
=======

.. autosummary::
    :toctree: generated/
    :nosignatures:


Functions
=========

.. autosummary::
    :toctree: generated/
    :nosignatures:

"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


# from .<module> import *
from .node_mixins import *  # noqa F403
from .edge_mixins import *  # noqa F403

__all__ = [name for name in dir() if not name.startswith('_')]



"""
********************************************************************************
compas_cem
********************************************************************************

.. currentmodule:: compas_cem


.. toctree::
    :maxdepth: 1

    compas_cem.diagrams
    compas_cem.elements
    compas_cem.loads
    compas_cem.supports
    compas_cem.equilibrium
    compas_cem.optimization
    compas_cem.plotters
"""

from __future__ import print_function

import os

__author__ = ["Rafael Pastrana"]
__copyright__ = "Copyright 2020 - Princeton University"
__license__ = "MIT License"
__email__ = "arpj@princeton.edu"
__version__ = "0.1.15"


# Directories
HERE = os.path.dirname(__file__)
HOME = os.path.abspath(os.path.join(HERE, "../../"))
DATA = os.path.abspath(os.path.join(HOME, "data"))
DOCS = os.path.abspath(os.path.join(HOME, "docs"))
TEMP = os.path.abspath(os.path.join(HOME, "temp"))
JSON_DATA = os.path.abspath(os.path.join(DATA, "json"))

# RGB color scheme
COLORS = {"tension": (227, 6, 75),  # red
          "compression": (12, 119, 184),  # blue
          "edge": (50, 50, 50),  # dark gray
          "node": (255, 255, 255),  # green
          "node_support": (0, 150, 10),  # green
          "node_origin": (155, 100, 255),  # purple
          "support_force": (0, 150, 10),  # green
          "load": (50, 250, 100),  # light green
          "trail": (250, 80, 210),  # pink
          "auxiliary_trail": (255, 155, 15)}  # orange

PROXY_PORT = 7123

__all_plugins__ = ['compas_cem.ghpython.install', 'compas_cem.ghpython.uninstall']
__all__ = ["HOME", "DATA", "DOCS", "TEMP", "JSON_DATA", "COLORS"]


# ==============================================================================
# Base element
# ==============================================================================

class Data(object):
    """
    Base class for all COMPAS data objects
    The base element.
    """
    def ToString(self):
        """
        Show object representation as string in Rhino/Grasshopper.
        """
        return self.__repr__()



"""
Make the length of a deviation edge to reach a goal length value.
"""
from ghpythonlib.componentbase import executingcomponent as component

from compas_cem.optimization import DeviationEdgeLengthConstraint


class DeviationEdgeLengthConstraintComponent(component):
    def RunScript(self, edge_key, length, weight):
        weight = weight or 1.0
        if edge_key and length is not None:
            return DeviationEdgeLengthConstraint(edge_key, length, weight)



"""
Set the force of a deviation edge as an optimization parameter.
"""
from ghpythonlib.componentbase import executingcomponent as component

from compas_cem.optimization import DeviationEdgeParameter


class DeviationEdgeParameterComponent(component):
    def RunScript(self, edge_key, bound_low, bound_up):
        if edge_key and bound_low is not None and bound_up is not None:
            return DeviationEdgeParameter(edge_key, bound_low, bound_up)



"""
Get the artists color scheme for the objects of the CEM framework.
"""
from ghpythonlib.componentbase import executingcomponent as component

from compas_cem import COLORS
import rhinoscriptsyntax as rs


class ArtistColorsComponent(component):
    def RunScript(self):
        colors = {key: rs.CreateColor(c) for key, c in COLORS.items()}

        color_tension = colors["tension"]
        color_compression = colors["compression"]
        color_edge = colors["edge"]
        color_node = colors["node"]
        color_node_support = colors["node_support"]
        color_node_origin = colors["node_origin"]
        color_load = colors["load"]
        color_reaction = colors["support_force"]
        color_trail = colors["trail"]
        color_auxiliary_trail = colors["auxiliary_trail"]

        colors = (color_tension, color_compression, color_edge, color_node, color_node_support, color_node_origin, color_load, color_reaction, color_trail, color_auxiliary_trail)

        return colors



"""
Set the length of a trail edge as an optimization parameter.
"""
from ghpythonlib.componentbase import executingcomponent as component

from compas_cem.optimization import TrailEdgeParameter


class TrailEdgeParameterComponent(component):
    def RunScript(self, edge_key, bound_low, bound_up):
        if edge_key and bound_low is not None and bound_up is not None:
            return TrailEdgeParameter(edge_key, bound_low, bound_up)



from compas.geometry import distance_point_point

from compas_cem.optimization.constraints import FloatConstraint


__all__ = ["DeviationEdgeLengthConstraint"]


class DeviationEdgeLengthConstraint(FloatConstraint):
    """
    Make a deviation edge reach a target length.
    """
    def __init__(self, edge=None, length=None, weight=1.0):
        # TODO: needs different serialization mechanism
        super(DeviationEdgeLengthConstraint, self).__init__(edge, length, weight)

    def reference(self, data):
        """
        """
        u, v = self.key()
        point_a = data["node_xyz"][u]
        point_b = data["node_xyz"][v]

        try:
            length = distance_point_point(point_a, point_b)
        except TypeError:
            # TODO: This import should not happen here
            import autograd.numpy as np
            length = np.linalg.norm(point_a - point_b)

        return length


if __name__ == "__main__":
    pass



from math import copysign

from compas.geometry import scale_vector
from compas.geometry import add_vectors
from compas.geometry import subtract_vectors
from compas.geometry import normalize_vector
from compas.geometry import length_vector
from compas.geometry import distance_point_point
from compas.geometry import intersection_line_plane

from compas_cem.diagrams import FormDiagram


__all__ = ["static_equilibrium"]


SMALL_VALUE = 1e-3


def static_equilibrium(topology, tmax=100, eta=1e-5, verbose=False, callback=None):
    """
    Generate a form diagram in static equilibrium.

    Parameters
    ----------
    topology : ``TopologyDiagram``
        A topology diagram
    tmax : ``int``. Optional.
        Maximum number of iterations the algorithm will run for.
        Defaults to ``100``.
    eta : ``float``. Optional.
        Distance threshold that marks equilibrium convergence.
        This threshold is compared against the sum of distances of the nodes'
        positions from one iteration to the next one.
        If ``eta`` is hit before consuming ``tmax`` iterations, calculations
        will stop early. Defaults to ``1e-5``.
    verbose : ``bool``. Optional.
        Flag to print out internal operations. Defaults to ``False``.
    callback : ``function``. Optional.
        An optional callback function to run at every iteration.

    Returns
    -------
    form : :class:`compas_cem.diagrams.FormDiagram`
        A form diagram.
    """
    attrs = equilibrium_state(topology, tmax, eta, verbose, callback)
    form = FormDiagram.from_topology_diagram(topology)
    form_update(form, **attrs)
    return form


def equilibrium_state(topology, tmax=100, eta=1e-5, verbose=False, callback=None):
    """
    Equilibrate forces at the nodes of a topology diagram.
    """
    trails = list(topology.trails())

    # there must be at least one trail
    assert len(trails) != 0, "No trails in the diagram!"

    # positions = {}
    residual_vectors = {}
    reaction_forces = {}
    trail_forces = {}
    node_xyz = {node: topology.node_coordinates(node) for node in topology.nodes()}

    for t in range(tmax):  # max iterations

        # store last positions for residual
        last_xyz = {k: v for k, v in node_xyz.items()}

        for i in topology.sequences():  # sequences

            for trail in trails:

                # if index is larger than available nodes in trails
                if i > (len(trail) - 1):
                    continue

                # select node from trail
                node = trail[i]

                # set initial trail vector and position for first iteration
                pos = node_xyz[node]
                if i == 0:
                    rvec = [0.0, 0.0, 0.0]

                # otherwise, select last trail vector and position from dictionary
                else:
                    rvec = residual_vectors[node]

                # calculate nodal equilibrium to get new residual vector
                indirect = True
                if t == 0:
                    indirect = False
                rvec = node_equilibrium(topology, node, rvec, node_xyz, indirect)

                # if this is the last node, store reaction force and exit loop
                if topology.is_node_support(node):
                    reaction_forces[node] = rvec[:]
                    continue

                # otherwise, pick next node in the trail
                next_node = trail[i + 1]

                # correct edge key
                edge = (node, next_node)
                if not topology.has_edge(*edge):
                    edge = (next_node, node)

                # query trail edge length
                length = topology.edge_attribute(key=edge, name="length")

                # query trail edge plane, it takes precedence over length
                plane = topology.edge_attribute(key=edge, name="plane")

                # override length if a plane exists
                if plane is not None:
                    # compute length from line plane intersection
                    p_length = trail_edge_length_from_plane(pos, rvec, plane)

                    # check that returned length is not null
                    if p_length is None:
                        print("Warning! No intersection found between edge {} and plane {}".format(edge, plane))
                        print("Falling back to input length: {}".format(length))
                        p_length = length
                        # raise ValueError("No intersection found between edge {} and plane {}".format(edge, plane))
                    # print out warning if zero length
                    elif p_length < SMALL_VALUE:
                        msg = "Warning! Length for edge after {} intersection is {}."
                        print(msg.format(edge, length), "Check the plane.")

                    # correct length sign based on internal force state
                    dot = p_length * length
                    if dot < 0:
                        length = p_length * -1.0
                    else:
                        length = p_length

                # store next node position
                next_pos = add_vectors(pos, scale_vector(normalize_vector(rvec), length))
                node_xyz[next_node] = next_pos

                # store trail force
                trail_forces[edge] = copysign(length_vector(rvec), length)

                # store residual vector
                residual_vectors[next_node] = rvec

                # do callback
                if callback:
                    callback()

        # if this is the first iteration, move directly to the next one
        if t == 0:
            continue

        # calculate residual
        residual = 0.0
        for key, pos in node_xyz.items():
            last_pos = last_xyz[key]
            residual += distance_point_point(last_pos, pos)

        # if residual smaller than threshold, stop iterating
        if residual < eta:
            break

    # if residual larger than threshold after tmax iterations, raise error
    if residual > eta:
        raise ValueError("Over {} iters. residual: {} > eta: {}".format(tmax, residual, eta))

    # print log
    if verbose:
        msg = "====== Completed Equilibrium in {} iters. Residual: {}======"
        print(msg.format(t, residual))

    eq_state = {}
    eq_state["node_xyz"] = node_xyz
    eq_state["trail_forces"] = trail_forces
    eq_state["reaction_forces"] = reaction_forces

    return eq_state


def form_update(form, node_xyz, trail_forces, reaction_forces):
    """
    Update the node and edge attributes of a form after equilibrating it.
    """
    # assign nodes' coordinates
    for node, xyz in node_xyz.items():
        form.node_xyz(node, xyz)
        # form.node_attributes(key=node, names=["x", "y", "z"], values=xyz)

    # assign forces on trail edges
    for edge, tforce in trail_forces.items():
        # tlength = form.edge_attribute(key=edge, name="length")
        # tforce = copysign(tforce, tlength)
        form.edge_attribute(key=edge, name="force", value=tforce)

    # assign reaction forces
    for node, rforce in reaction_forces.items():
        form.node_attributes(key=node, names=["rx", "ry", "rz"], values=rforce)

    # assign lengths to deviation edges
    for u, v in form.edges():
        length = form.edge_length(u, v)
        force = form.edge_attribute(key=edge, name="force")
        length = copysign(length, force)
        form.edge_attribute(key=(u, v), name="length", value=length)


def node_equilibrium(form, node, t_vec, node_xyz, indirect=False, verbose=False):
    """
    Calculates the equilibrium of trail and deviation forces at a node.

    Parameters
    ----------
    form : ``FormDiagram``
        A form diagram
    node : ``int``
        A node key.
    t_vec : ``list``
        A trail vector.
    indirect : ``bool``
        Flag to consider indirect deviation edges in the calculation.
        Defaults to ``False``.
    verbose : ``bool``
        Flag to print out internal output. Defaults to ``False``.

    Returns
    -------
    t_vec : ``list``
        The new trail vector.
    """
    tvec_in = scale_vector(t_vec, -1.0)
    q_vec = form.node_load(node)
    rd_vec = direct_deviation_edges_resultant_vector(form, node, node_xyz)

    if indirect:
        ri_vec = indirect_deviation_edges_resultant_vector(form, node, node_xyz)
    else:
        ri_vec = [0.0, 0.0, 0.0]

    tvec_out = trail_vector_out(tvec_in, q_vec, rd_vec, ri_vec)

    if verbose:
        print("-----")
        print("node", node)
        print("qvec", q_vec)
        print("rd_vec", rd_vec)
        print("ri_vec", rd_vec)
        print("t vec np", tvec_out)

    return tvec_out


def incoming_edge_vectors(node, node_xyz, edges, normalize=False):
    """
    Temporary alternative to Diagram.incoming_edge_vectors()
    """
    vectors = []
    for u, v in edges:
        other = u if u != node else v
        vector = vector_two_nodes(node_xyz[other], node_xyz[node], normalize)
        vectors.append(vector)

    return vectors


def vector_two_nodes(a, b, normalize=False):
    """
    Calculates the vector between the xyz coordinates of two noddes.

    Parameters
    ----------
    a : ``list``
        The xyz coordinates of the first node
    b : ``list``
        The xyz coordinates of the second node
    normalize : ``bool``
        A boolean flag to normalize all the resulting edge vectors.
        Defaults to ``False``.

    Returns
    -------
    vector : ``list``
        The calculated xyz vector.
    """
    vector = subtract_vectors(a, b)
    if not normalize:
        return vector
    return normalize_vector(vector)


def deviation_edges_resultant_vector(form, node, node_xyz, deviation_edges):
    """
    Adds up the force vectors of the deviation edges incident to a node.

    Parameters
    ----------
    form : ``FormDiagram``
        A form diagram.
    node : ``int``
        A node key.
    node_xyz : ``dict``
        A dictionary with node keys and xyz coordinates as values.
    deviation_edges : ``list``
        A list with deviation edges keys.

    Returns
    -------
    rvec : ``list``
        The resulting force vector.
    """
    r_vec = [0.0, 0.0, 0.0]
    if not deviation_edges:
        return r_vec

    # vectors = form.incoming_edge_vectors(node, deviation_edges, True)
    vectors = incoming_edge_vectors(node, node_xyz, deviation_edges, True)
    forces = form.edges_attribute(name="force", keys=deviation_edges)

    for force, dev_vec in zip(forces, vectors):
        r_vec = add_vectors(r_vec, scale_vector(dev_vec, force))

    return r_vec


def direct_deviation_edges_resultant_vector(form, node, node_xyz):
    """
    Adds up the force vectors of the direct deviation edges incident to a node.

    Parameters
    ----------
    form : ``FormDiagram``
        A form diagram.
    node : ``int``
        A node key.

    Returns
    -------
    rvec : ``list``
        The resulting force vector.
    """
    deviation_edges = form._connected_direct_deviation_edges(node)
    return deviation_edges_resultant_vector(form, node, node_xyz, deviation_edges)


def indirect_deviation_edges_resultant_vector(form, node, node_xyz):
    """
    Adds up the force vectors of the indirect deviation edges incident to a node.

    Parameters
    ----------
    form : ``FormDiagram``
        A form diagram.
    node : ``int``
        A node key.

    Returns
    -------
    rvec : ``list``
        The resulting force vector.
    """
    deviation_edges = form._connected_indirect_deviation_edges(node)
    return deviation_edges_resultant_vector(form, node, node_xyz, deviation_edges)


def trail_vector_out(tvec_in, q_vec, rd_vec, ri_vec):
    """
    Calculate an outgoing trail vector.

    Parameters
    ----------
    tvec_in : ``list``
        An incoming trail vector.
    q_vec : ``list``
        A load vector.
    rd_vec : ``list``
        The direct deviation edges resultant force vector.
    ri_vec : ``list``
        The indirect deviation edges resultant force vector.

    Returns
    -------
    tvec_out : ``list``
        An outgoing trail vector.
    """
    tvec = [0.0, 0.0, 0.0]
    vectors = [tvec_in, q_vec, rd_vec, ri_vec]
    for vec in vectors:
        tvec = add_vectors(tvec, vec)
    return scale_vector(tvec, -1.0)


def trail_edge_length_from_plane(pos, direction, plane):
    """
    Calculates the length of a trail edge from a line-plane intersection.

    Parameters
    ----------
    pos : ``list`` of ``float``
        The XYZ coordinates of the starting node of an edge.
    direction : ``list`` of ``float``
        The XYZ coordinates of the direction the edge points to.
    plane : ``Plane``
        A COMPAS plane.

    Returns
    -------
    length : ``float``
        The distance between ``pos`` and the resulting line-plane intersection.
        If not intersection is found, it returns ``None``.

    Notes
    -----
    A line will be created by adding ``pos`` and ``direction``.
    The line functions as a line ray, not as finite-length segment.
    """
    # compute length from line plane intersection
    line = (pos, add_vectors(pos, direction))
    pos_plane = intersection_line_plane(line, plane)
    if pos_plane:
        return distance_point_point(pos, pos_plane)


if __name__ == "__main__":
    pass



"""
Draw a topology diagram.
"""
from ghpythonlib.componentbase import executingcomponent as component

from compas_cem.ghpython import TopologyArtist


class TopologyArtistComponent(component):
    def RunScript(self, topology, node_keys, edge_keys, force_min, force_scale):
        node_keys = node_keys or None
        edge_keys = edge_keys or None
        force_min = force_min or 1e-3
        force_scale = force_scale or 1.0
        if topology:
            artist = TopologyArtist(topology)

            nodes = artist.draw_nodes(node_keys)
            origin_nodes = artist.draw_nodes_origin(node_keys)
            support_nodes = artist.draw_nodes_support(node_keys)

            edges = artist.draw_edges(edge_keys)
            trail_edges = artist.draw_edges_trail(edge_keys)
            deviation_edges = artist.draw_edges_deviation(edge_keys)

            trails = artist.draw_trails()

            loads = artist.draw_loads(node_keys, min_load=force_min, scale=force_scale)

            return nodes, origin_nodes, support_nodes, edges, trail_edges, deviation_edges, trails, loads



from time import time


__all__ = ["solve_nlopt_proxy"]


# ------------------------------------------------------------------------------
# Optimization
# ------------------------------------------------------------------------------

def solve_nlopt_proxy(topology, constraints, parameters, algorithm, iters, eps, tmax=100, eta=1e-6):
    """
    Solve a constrained form-finding task through a Proxy server.
    """
    from compas_cem.optimization import Optimizer

    optimizer = Optimizer()

    # add constraints
    for constraint in constraints:
        optimizer.add_constraint(constraint)

    # add parameters
    for parameter in parameters:
        optimizer.add_parameter(parameter)

    start = time()
    form = optimizer.solve_nlopt(topology, algorithm, iters, eps, tmax, eta)

    duration = time() - start
    objective = optimizer.penalty
    evals = optimizer.evals
    grad_norm = optimizer.gradient_norm
    status = optimizer.status

    return form, objective, grad_norm, evals, duration, status

# ------------------------------------------------------------------------------
# Main
# ------------------------------------------------------------------------------


if __name__ == "__main__":
    pass



"""
Create a trail edge from a rhino line.
"""
from ghpythonlib.componentbase import executingcomponent as component

from compas.geometry import Plane
from compas.geometry import cross_vectors

from compas_rhino.geometry import RhinoPlane

from compas_cem.elements import TrailEdge


class TrailEdgeComponent(component):
    def RunScript(self, line, length, plane):
        if plane is not None:
            plane = RhinoPlane.from_geometry(plane).to_compas()
            plane = Plane(plane.point, cross_vectors(plane.xaxis, plane.yaxis))
        if line and length is not None:
            return TrailEdge.from_rhino_line(line, length=length, plane=plane)



from math import fabs

from compas_cem.optimization import Serializable


__all__ = ["OriginNodeXParameter",
           "OriginNodeYParameter",
           "OriginNodeZParameter"]

# ------------------------------------------------------------------------------
# Base Node Parameter
# ------------------------------------------------------------------------------


class NodeParameter(Serializable):
    """
    Parametrize a node attribute to solve an optimization problem.
    """
    def __init__(self, key, bound_low, bound_up):
        self._key = key
        self._bound_up = bound_up
        self._bound_low = bound_low
        self._attr_name = None

    def key(self):
        """
        The node key.
        """
        return self._key

    def start_value(self, topology):
        """
        The initial value of the node optimization parameter.
        """
        val = topology.node_attribute(key=self.key(), name=self._attr_name)
        return val

    def bound_low(self, topology):
        """
        The lower bound of the node optimization parameter.

        Notes
        -----
        Calculated as the initial parameter minus this bound's absolute value.
        """
        return self.start_value(topology) - fabs(self._bound_low)

    def bound_up(self, topology):
        """
        The upper bound of the node optimization parameter.

        Notes
        -----
        Calculated as the initial parameter plus this bound's absolute value.
        """
        return self.start_value(topology) + fabs(self._bound_up)

    def attr_name(self):
        """
        The name of the node attribute to parametrize.
        """
        return self._attr_name

    def __repr__(self):
        """
        """
        st = "{0}(key={1!r}, bound_low={2!r}, bound_up={3!r})"
        return st.format(self.__class__.__name__, self._key, self._bound_low, self._bound_up)

# ------------------------------------------------------------------------------
# Data
# ------------------------------------------------------------------------------

    @property
    def data(self):
        """
        A data dictionary that represents an ``NodeParameter`` object.

        Returns
        -------
            data : ``dict``
            A dictionary that contains the following key-value pairs:

            * "key" : ``tuple``
            * "bound up" : ``float``
            * "bound low" : ``float``
            * "attr name" : ``str``

        Notes
        -----
        All dictionary keys are converted to their representation value
        (``repr(key)``) to ensure compatibility of all allowed key types with
        the JSON serialization format, which only allows for dict keys that are strings.
        """

        # TODO: Check serialization mechanism.

        data = {}

        data["key"] = self._key
        data["bound_up"] = self._bound_up
        data["bound_low"] = self._bound_low
        data["attr_name"] = self._attr_name
        data["datatype"] = self.datatype()

        return data

    @data.setter
    def data(self, data):
        """
        Overwrites this object's attributes with a data dictionary.

        Parameters
        ----------
        data : ``dict``
            A data dictionary.
        """
        self._key = int(data["key"])
        self._bound_up = float(data["bound_up"])
        self._bound_low = float(data["bound_low"])
        self._attr_name = str(data["attr_name"])

# ------------------------------------------------------------------------------
# Origin Node Constraint on X
# ------------------------------------------------------------------------------


class OriginNodeXParameter(NodeParameter):
    """
    Sets the X coordinate of an origin node as an optimization parameter.
    """
    def __init__(self, key=None, bound_low=None, bound_up=None):
        super(OriginNodeXParameter, self).__init__(key, bound_low, bound_up)
        self._attr_name = "x"

# ------------------------------------------------------------------------------
# Origin Node Constraint on Y
# ------------------------------------------------------------------------------


class OriginNodeYParameter(NodeParameter):
    """
    Sets the Y coordinate of an origin node as an optimization parameter.
    """
    def __init__(self, key=None, bound_low=None, bound_up=None):
        super(OriginNodeYParameter, self).__init__(key, bound_low, bound_up)
        self._attr_name = "y"

# ------------------------------------------------------------------------------
# Origin Node Constraint on Z
# ------------------------------------------------------------------------------


class OriginNodeZParameter(NodeParameter):
    """
    Sets the Z coordinate of an origin node as an optimization parameter.
    """
    def __init__(self, key=None, bound_low=None, bound_up=None):
        super(OriginNodeZParameter, self).__init__(key, bound_low, bound_up)
        self._attr_name = "z"


if __name__ == "__main__":

    from time import time

    from compas.geometry import Point

    from compas_cem.diagrams import TopologyDiagram

    from compas_cem.elements import Node
    from compas_cem.elements import TrailEdge
    from compas_cem.elements import DeviationEdge

    from compas_cem.loads import NodeLoad

    from compas_cem.supports import NodeSupport

    from compas_cem.optimization import Optimizer
    from compas_cem.optimization import DeviationEdgeParameter
    from compas_cem.optimization import TrailEdgeParameter

    from compas_cem.optimization import PointConstraint

    from compas_cem.equilibrium import static_equilibrium

    from compas_cem.plotters import FormPlotter

    # create a topology diagram
    topology = TopologyDiagram()

    # add nodes
    topology.add_node(Node(0, [0.0, 0.0, 0.0]))
    topology.add_node(Node(1, [1.0, 0.0, 0.0]))
    topology.add_node(Node(2, [2.0, 0.0, 0.0]))
    topology.add_node(Node(3, [3.0, 0.0, 0.0]))

    # add edges with negative values for a compression-only structure
    topology.add_edge(TrailEdge(0, 1, length=-1.0))
    topology.add_edge(DeviationEdge(1, 2, force=-1.0))
    topology.add_edge(TrailEdge(2, 3, length=-1.0))

    # add supports
    topology.add_support(NodeSupport(0))
    topology.add_support(NodeSupport(3))

    # add loads
    topology.add_load(NodeLoad(1, [0.0, -1.0, 0.0]))
    topology.add_load(NodeLoad(2, [0.0, -1.0, 0.0]))

    # calculate equilibrium
    topology.build_trails()
    form = static_equilibrium(topology, eta=1e-5, tmax=100)

    # optimization
    optimizer = Optimizer()

    optimizer.add_parameter(OriginNodeYParameter(1, 1.0, 1.0))
    optimizer.add_parameter(DeviationEdgeParameter((1, 2), 1.0, 1.0))
    optimizer.add_parameter(TrailEdgeParameter((2, 3), 1.0, 1.0))

    point_a = Point(1.0, -0.5, 0.0)
    optimizer.add_constraint((PointConstraint(1, point_a)))

    point_b = Point(3.0, -0.707, 0.0)
    optimizer.add_constraint((PointConstraint(3, point_b)))

    # optimization settings
    start = time()
    algo = "LD_LBFGS"  # LN_BOBYQA, LD_LBFGS, LD_MMA
    iters = 100  # 100
    eps = 1e-6  # 1e-4

    # optimize
    form = optimizer.solve_nlopt(topology, algo, iters, eps)

    print("Elapsed time: {}".format(time() - start))
    print("Total error: {}".format(optimizer.penalty))

    # plot
    plotter = FormPlotter(topology, figsize=(16, 9))

    plotter.draw_nodes(radius=0.03, text="key-xyz")
    plotter.draw_edges(text="force-length")
    plotter.draw_loads(scale=-0.25)
    plotter.draw_reactions(scale=0.10)
    plotter.show()



from compas.geometry import closest_point_on_plane

from compas_cem.optimization.constraints import VectorConstraint


__all__ = ["PlaneConstraint"]


class PlaneConstraint(VectorConstraint):
    """
    Pulls the xyz position of a node to a target plane.
    """
    def __init__(self, node=None, plane=None, weight=1.0):
        super(PlaneConstraint, self).__init__(node, plane, weight)

    def reference(self, data):
        """
        """
        return data["node_xyz"][self.key()]

    def target(self, point):
        """
        """
        plane = self._target
        return closest_point_on_plane(point, plane)


if __name__ == "__main__":
    pass



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import glob
import os

import compas
from compas_ghpython.components import uninstall_userobjects


@compas.plugins.plugin(category='install')
def after_rhino_uninstall(uninstalled_packages):
    if 'compas_cem' not in uninstalled_packages:
        return []
    srcdir = os.path.join(os.path.dirname(__file__), 'components', 'ghuser')
    userobjects = [os.path.basename(ghuser) for ghuser in glob.glob(os.path.join(srcdir, '*.ghuser'))]
    uninstalled_objects = uninstall_userobjects(userobjects)

    uninstall_errors = [uo[0] for uo in uninstalled_objects if not uo[1]]
    error_msg = '' if not uninstall_errors else 'and {} failed to uninstall'.format(len(uninstall_errors))

    return [('compas_cem', 'Uninstalled {} GH User Objects {}'.format(len(uninstalled_objects), error_msg), True)]



"""
Pull the position of a node to a target line ray.
"""
from ghpythonlib.componentbase import executingcomponent as component

from compas_cem.optimization import LineConstraint
from compas_rhino.geometry import RhinoLine


class LineConstraintComponent(component):
    def RunScript(self, node_key, line, weight):
        weight = weight or 1.0
        if node_key is not None and line:
            line = RhinoLine.from_geometry(line).to_compas()
            return LineConstraint(node_key, line, weight)



"""
Export a COMPAS CEM diagram and save it as a JSON file.
"""
import os

from ghpythonlib.componentbase import executingcomponent as component


class DiagramToJSON(component):
    def RunScript(self, diagram, filepath):
        if diagram and filepath:
            filepath = os.path.abspath(filepath)
            diagram.to_json(filepath, pretty=True)
            return diagram.data



"""
Pull the position of a node to a target plane.
"""
from ghpythonlib.componentbase import executingcomponent as component

from compas.geometry import Plane
from compas.geometry import cross_vectors

from compas_rhino.geometry import RhinoPlane
from compas_cem.optimization import PlaneConstraint


class PlaneConstraintComponent(component):
    def RunScript(self, node_key, plane, weight):
        weight = weight or 1.0
        if node_key is not None and plane is not None:
            plane = RhinoPlane.from_geometry(plane).to_compas()
            plane = Plane(plane.point, cross_vectors(plane.xaxis, plane.yaxis))
            return PlaneConstraint(node_key, plane, weight)



"""
Make the reaction force on a trail edge to meet a target force vector.
"""
from ghpythonlib.componentbase import executingcomponent as component

from compas_rhino.geometry import RhinoVector
from compas_cem.optimization import ReactionForceConstraint


class ReactionForceConstraintComponent(component):
    def RunScript(self, node_key, vector, weight):
        weight = weight or 1.0
        if node_key is not None and vector:
            vector = RhinoVector.from_geometry(vector).to_compas()
            return ReactionForceConstraint(node_key, vector, weight)



"""
compas_cem.optimization
****************************

.. currentmodule:: compas_cem.optimization

Optimizers
==========

.. autosummary::
    :toctree: generated/
    :nosignatures:

    Optimizer
    solve_nlopt_proxy

Constraints
===========

.. autosummary::
    :toctree: generated/
    :nosignatures:

    PointConstraint
    LineConstraint
    PlaneConstraint
    DeviationEdgeLengthConstraint
    TrailEdgeForceConstraint
    ReactionForceConstraint

Optimization Parameters
=======================

.. autosummary::
    :toctree: generated/
    :nosignatures:

    TrailEdgeParameter
    DeviationEdgeParameter
    OriginNodeXParameter
    OriginNodeYParameter
    OriginNodeZParameter

NLOpt
=====
.. autosummary::
    :toctree: generated/
    :nosignatures:

    nlopt_algorithm
    nlopt_algorithms
    nlopt_solver

"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


# from .<module> import *
from ._serialization import *  # noqa F403
from .constraints import *  # noqa F403
from .parameters import *  # noqa F403
from .proxy import *  # noqa F403

import compas
if not compas.IPY:
    from .nlopt import *  # noqa F403
    from .objective_func import *  # noqa F403
    from .grad import *  # noqa F403
    from .optimizer import *  # noqa F403


__all__ = [name for name in dir() if not name.startswith('_')]



from compas_cem.elements import Edge


class TrailEdge(Edge):
    """
    A trail edge.

    Notes
    -----
    If a plane is defined, it will override the absolute length of the trail edge.
    However, the sign of the length (e.g. the combinatorial state) is preserved.

    TODO: addexplicit combinatorial state to the signature of the constructor.
    """
    def __init__(self, u, v, length, plane=None):
        super(TrailEdge, self).__init__(u, v)
        # TODO
        # self.attributes = {"length": length, "state": state, type": "trail", "plane": plane}
        self.attributes = {"length": length, "type": "trail", "plane": plane}

    def __repr__(self):
        """
        """
        st = "{0}(length={1!r}, plane={2!r})"
        return st.format(self.__class__.__name__, self.attributes["length"], self.attributes["plane"])

# ==============================================================================
# Main
# ==============================================================================


if __name__ == "__main__":
    pass



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os

import compas.plugins
from compas_ghpython.components import install_userobjects


@compas.plugins.plugin(category='install')
def installable_rhino_packages(category='install'):
    return ['compas_cem']


@compas.plugins.plugin(category='install')
def after_rhino_install(installed_packages):
    if 'compas_cem' not in installed_packages:
        return []

    installed_objects = install_userobjects(os.path.join(os.path.dirname(__file__), 'components', 'ghuser'))

    return [('compas_cem', 'Installed {} GH User Objects'.format(len(installed_objects)), True)]



"""
Search for an edge key in a topology or a form diagram using a line.
"""
from ghpythonlib.componentbase import executingcomponent as component

from compas.utilities import geometric_key
from compas_rhino.geometry import RhinoLine


class EdgeSearchComponent(component):
    def RunScript(self, diagram, line):
        if diagram and line:

            # edge = []
            # for ln in line:
            ln = RhinoLine.from_geometry(line).to_compas()

            eg = []
            for point in (ln.start, ln.end):
                gkey = geometric_key(point, diagram.tol)
                node = diagram.gkey_node[gkey]
                eg.append(node)
            # edge.append(tuple(eg))

            return tuple(eg)



"""
Extract the post form-finding results of the edges in a form diagram.
"""
from ghpythonlib.componentbase import executingcomponent as component


class EdgeResultsComponent(component):
    def RunScript(self, form, edge_keys):
        if form:
            edge_keys = edge_keys or list(form.edges())
            lengths = [form.edge_length_2(ed) for ed in edge_keys]
            forces = [form.edge_force(ed) for ed in edge_keys]
            return lengths, forces



"""
Create a deviation edge from a rhino line.
"""
from ghpythonlib.componentbase import executingcomponent as component

from compas_cem.elements import DeviationEdge


class DeviationEdgeComponent(component):
    def RunScript(self, line, force):
        if line and force is not None:
            return DeviationEdge.from_rhino_line(line, force=force)



"""
Create a node support from a Rhino point.
"""
from ghpythonlib.componentbase import executingcomponent as component

from compas_cem.supports import NodeSupport


class NodeSupportComponent(component):
    def RunScript(self, point):
        if point:
            return NodeSupport.from_rhino_point(point)



"""
compas_cem.ghpython
****************************

.. currentmodule:: compas_cem.ghpython


Artists
========

In **COMPAS CEM**, the `artists` are classes that assist with the visualization
of diagrams, in a way that maintains the data separated from the specific CAD
interfaces, while providing a way to leverage native performance of the CAD
environment.

.. autosummary::
    :toctree: generated/
    :nosignatures:

    TopologyArtist
    FormArtist
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import compas

if compas.RHINO:
    from .artists import *  # noqa F403

__all__ = [name for name in dir() if not name.startswith('_')]



from compas_cem.optimization.constraints import FloatConstraint
from compas_cem.optimization.constraints import VectorConstraint


__all__ = ["TrailEdgeForceConstraint",
           "ReactionForceConstraint"]


class TrailEdgeForceConstraint(FloatConstraint):
    """
    Make a trail edge reach a target force value.
    """
    def __init__(self, edge=None, force=None, weight=1.0):
        super(TrailEdgeForceConstraint, self).__init__(edge, force, weight)

    def reference(self, data):
        """
        """
        return data["trail_forces"][self.key()]


class ReactionForceConstraint(VectorConstraint):
    """
    Makes the support reaction force at a node match a target vector.
    """
    def __init__(self, node=None, vector=None, weight=1.0):
        super(ReactionForceConstraint, self).__init__(node, vector, weight)

    def reference(self, data):
        """
        """
        return data["reaction_forces"][self.key()]


if __name__ == "__main__":

    from time import time

    from compas.geometry import Point

    from compas_cem.diagrams import TopologyDiagram

    from compas_cem.elements import Node
    from compas_cem.elements import TrailEdge
    from compas_cem.elements import DeviationEdge

    from compas_cem.loads import NodeLoad

    from compas_cem.supports import NodeSupport

    from compas_cem.optimization import Optimizer
    from compas_cem.optimization import DeviationEdgeParameter

    from compas_cem.optimization import OriginNodeXParameter
    from compas_cem.optimization import OriginNodeYParameter

    from compas_cem.optimization import PointConstraint

    from compas_cem.equilibrium import static_equilibrium

    from compas_cem.plotters import FormPlotter

    # create a topology diagram
    topology = TopologyDiagram()

    # add nodes
    topology.add_node(Node(0, [0.0, 0.0, 0.0]))
    topology.add_node(Node(1, [1.0, 0.0, 0.0]))
    topology.add_node(Node(2, [2.0, 0.0, 0.0]))
    topology.add_node(Node(3, [0.0, 1.0, 0.0]))
    topology.add_node(Node(4, [1.0, 1.0, 0.0]))
    topology.add_node(Node(5, [2.0, 1.0, 0.0]))

    # add edges with negative values for a compression-only structure
    topology.add_edge(TrailEdge(0, 3, length=-1.0))
    topology.add_edge(TrailEdge(1, 4, length=-1.0))
    topology.add_edge(TrailEdge(2, 5, length=-1.0))

    topology.add_edge(DeviationEdge(3, 4, force=-1.0))
    topology.add_edge(DeviationEdge(4, 5, force=-1.0))

    # add supports
    topology.add_support(NodeSupport(0))
    topology.add_support(NodeSupport(1))
    topology.add_support(NodeSupport(2))

    # add loads
    topology.add_load(NodeLoad(4, [0.0, -1.0, 0.0]))

    # calculate equilibrium
    topology.build_trails()
    form = static_equilibrium(topology)

    # optimization
    optimizer = Optimizer()

    # root node parameter in x and y
    for node in (3, 4, 5):
        optimizer.add_parameter(OriginNodeXParameter(node, 1.0, 1.0))
        optimizer.add_parameter(OriginNodeYParameter(node, 1.0, 1.0))

    # deviation edge parameters
    optimizer.add_parameter(DeviationEdgeParameter((3, 4), 1.0, 1.0))
    optimizer.add_parameter(DeviationEdgeParameter((4, 5), 1.0, 1.0))

    # goals
    point_a = Point(0.0, 1.5, 0.0)
    optimizer.add_constraint((PointConstraint(3, point_a)))
    optimizer.add_constraint(TrailEdgeForceConstraint((1, 4), 0.0))
    optimizer.add_constraint(ReactionForceConstraint(1, [0.0, 0.0, 0.0]))

    # optimization settings
    start = time()
    algo = "LD_LBFGS"  # LN_BOBYQA, LD_LBFGS, LD_MMA
    iters = 100  # 100
    eps = 1e-6  # 1e-4

    # optimize
    form = optimizer.solve_nlopt(topology, algo, iters, eps)

    print("Elapsed time: {}".format(time() - start))
    print("Total error: {}".format(optimizer.penalty))

    # plot
    plotter = FormPlotter(form, figsize=(16, 9))

    plotter.draw_nodes(radius=0.03, text="key-xyz")
    plotter.draw_loads(scale=-0.25)
    plotter.draw_reactions(scale=1.0)
    plotter.draw_edges(text="force")
    plotter.show()



"""
Generate a form diagram in static equilibrium.
"""
from ghpythonlib.componentbase import executingcomponent as component

from compas_cem.equilibrium import static_equilibrium


class FormFindingComponent(component):
    def RunScript(self, topology, tmax, eta):
        tmax = tmax or 100
        eta = eta or 1e-6
        if topology:
            return static_equilibrium(topology, tmax=tmax, eta=eta)



"""
Pull the position of a node to a target point.
"""
from ghpythonlib.componentbase import executingcomponent as component

from compas_cem.optimization import PointConstraint
from compas_rhino.geometry import RhinoPoint


class PointConstraintComponent(component):
    def RunScript(self, node_key, point, weight):
        weight = weight or 1.0
        if node_key is not None and point:
            point = RhinoPoint.from_geometry(point).to_compas()
            return PointConstraint(node_key, point, weight)



from abc import abstractmethod
from ast import literal_eval

from compas.geometry import distance_point_point_sqrd

from compas_cem.optimization import Serializable


# ------------------------------------------------------------------------------
# Constraint
# ------------------------------------------------------------------------------


class Constraint(Serializable):
    """
    The blueprint of a constraint.
    """
    def __init__(self, key, target, weight):
        self._key = key  # a topological key
        self._target = target  # a geometric target
        self._weight = weight  # the strength of the constraint

    def key(self):
        """
        The key to an edge or a node in a form diagram.

        Returns
        -------
        key : ``int`` or ``tuple``
            The key to a node of an edge.
        """
        return self._key

    def target(self, *args, **kwargs):
        """
        The target to reach.

        Returns
        -------
        target : ``object``
            A target object.
        args : ``list``
            A list of additional arguments.
        kwargs : ``dict``
            A dictionary with extra named arguments
        """
        return self._target

    @property
    def weight(self):
        """
        The strength of the constraint.
        """
        return self._weight

    @abstractmethod
    def penalty(self):
        """
        Calculate the penalty caused by the constraint.
        """
        return

    def __repr__(self):
        st = "{0}(key={1!r}, target={2!r}, weight={3!r})"
        return st.format(self.__class__.__name__, self._key, self._target, self._weight)

# ------------------------------------------------------------------------------
# Vector Constraint
# ------------------------------------------------------------------------------


class VectorConstraint(Constraint):
    """
    The blueprint of a constraint that measures distances between two vectors.
    """
    def penalty(self, data):
        """
        The distance between the current and the target vector.

        Returns
        -------
        error : ``float``
            The squared difference.
        """
        vec_a = self.reference(data)
        vec_b = self.target(vec_a)

        return distance_point_point_sqrd(vec_a, vec_b) * self.weight

    @property
    def data(self):
        """
        A data dictionary that represents a ``Constraint`` object.

        Returns
        -------
        data : ``dict``
            A dictionary that contains the following key-value pairs:

            * "key" : ``str``
            * "datatype" : ``str`
            * "target" : ``str``
            * "target_datatype" : ``str``
        """
        data = {}

        data["key"] = repr(self.key())
        data["target"] = self._target.to_data()
        data["datatype"] = self.datatype()
        data["target_datatype"] = self.object_datatype(self._target)

        return data

    @data.setter
    def data(self, data):
        """
        Overwrites this object's attributes with a data dictionary.

        Parameters
        ----------
        data : ``dict``
            A dictionary that contains the following key-value pairs:

            * "key" : ``str``
            * "datatype" : ``str`
            * "target" : ``str``
            * "target_datatype" : ``str``
        """
        self._key = literal_eval(data["key"])
        target_cls = self.object_cls_from_dtype(data["target_datatype"])
        target = target_cls.from_data(data["target"])
        self._target = target

# ------------------------------------------------------------------------------
# Float Constraint
# ------------------------------------------------------------------------------


class FloatConstraint(Constraint):
    """
    The blueprint of a constraint that measures distances between two floats.
    """
    def penalty(self, data):
        """
        The distance between the current and the target float.

        Returns
        -------
        error : ``float``
            The squared difference.
        """
        float_a = self.reference(data)
        float_b = self.target(float_a)
        diff = float_a - float_b

        return diff * diff * self.weight

    @property
    def data(self):
        """
        A data dictionary that represents a ``Constraint`` object.

        Returns
        -------
        data : ``dict``
            A dictionary that contains the following key-value pairs:

            * "key" : ``str``
            * "datatype" : ``str`
            * "target" : ``str``
        """
        data = {}

        data["key"] = repr(self.key())
        data["target"] = repr(self.target())
        data["datatype"] = self.datatype()

        return data

    @data.setter
    def data(self, data):
        """
        Overwrites this object's attributes with a data dictionary.

        Parameters
        ----------
        data : ``dict``
            A dictionary that contains the following key-value pairs:

            * "key" : ``str``
            * "target" : ``str`
        """
        self._key = literal_eval(data["key"])
        self._target = literal_eval(data["target"])

# ==============================================================================
# Main
# ==============================================================================


if __name__ == "__main__":
    pass



"""
Import a COMPAS CEM form diagram from a JSON file.
"""
import os

from ghpythonlib.componentbase import executingcomponent as component

from compas_cem.diagrams import FormDiagram


class FormDiagramFromJSON(component):
    def RunScript(self, filepath):
        if filepath:
            filepath = os.path.abspath(filepath)
            form = FormDiagram.from_json(filepath)
            return form



"""
Set the Z coordinate of an origin node as an optimization parameter.
"""
from ghpythonlib.componentbase import executingcomponent as component

from compas_cem.optimization import OriginNodeZParameter


class OriginNodeZParameterComponent(component):
    def RunScript(self, node_key, bound_low, bound_up):
        if node_key is not None and bound_low is not None and bound_up is not None:
            return OriginNodeZParameter(node_key, bound_low, bound_up)



"""
Create a load vector to be applied at a node.
"""
from ghpythonlib.componentbase import executingcomponent as component

from compas_cem.loads import NodeLoad


class NodeLoadComponent(component):
    def RunScript(self, point, vector):
        if point and vector:
            return NodeLoad.from_rhino_point_and_vector(point, vector)



import compas

if compas.RHINO:
    from compas_rhino.geometry import RhinoPoint

from compas_cem import Data


__all__ = ["NodeSupport"]

# ==============================================================================
# Node Support
# ==============================================================================


class NodeSupport(Data):
    """
    A support assigned to a node.

    Parameters
    ----------
    node : ``int``
        The key of the node where to apply the support to.

    Returns
    -------
    node_support : ``NodeSupport``
        A node support.
    """
    def __init__(self, node):
        self.node = node
        self.xyz = None

    @classmethod
    def from_point(cls, point):
        """
        Create a NodeSupport from a point.

        Parameters
        ----------
        point : ``list`` of ``float``
            The xyz coordinates of the positions where to assign a support.

        Returns
        -------
        support : ``NodeSupport``
            A support object.

        Notes
        -----
        The support will be assigned to a ``FormDiagram`` only if it has a node
        whose xyz coordinates matches defined by the ``NodeSupport``.
        Otherwise, the support will not be assigned to the diagram.
        """
        support = cls(node=None)
        support.xyz = point
        return support

    @classmethod
    def from_rhino_point(cls, rhino_point):
        """
        Create a NodeSupport from a rhino point.

        Parameters
        ----------
        point : ``Rhino.Geometry.Point3d``
            The xyz position where to apply a support.

        Returns
        -------
        support : ``NodeSupport``
            A support object.

        Notes
        -----
        The support will be assigned to a ``FormDiagram`` only if it has a node
        whose xyz coordinates matches defined by the ``NodeSupport``.
        Otherwise, the support will not be assigned to the diagram.
        """
        xyz = RhinoPoint.from_geometry(rhino_point).to_compas()
        return cls.from_point(xyz)

    def __repr__(self):
        """
        """
        return "{0}(xyz={1!r})".format(self.__class__.__name__, self.xyz)

# ==============================================================================
# Main
# ==============================================================================


if __name__ == "__main__":
    pass



from compas.geometry import scale_vector
from compas.geometry import add_vectors
from compas.geometry import normalize_vector

from compas_cem.diagrams import Diagram

from compas_cem.elements import Node
from compas_cem.elements import TrailEdge

from compas_cem.supports import NodeSupport


__all__ = ["TopologyDiagram"]

# ==============================================================================
# Form Diagram
# ==============================================================================


class TopologyDiagram(Diagram):
    """
    The very heart of life.

    Parameters
    ----------
    *args : ``list``
        Arguments.
    **kwargs : ``dict``
        Keyword arguments.

    Returns
    -------
    topology : ``TopologyDiagram``
        A topology diagram.
    """

    def __init__(self, *args, **kwargs):
        super(TopologyDiagram, self).__init__(*args, **kwargs)

        self.attributes["_trails"] = dict()
        self.attributes["_auxiliary_trails"] = dict()
        self.attributes["_aux_length"] = -1.0
        self.attributes["_aux_vector"] = [1.0, 1.0, 1.0]

# ==============================================================================
# Properties
# ==============================================================================

    @property
    def auxiliary_trail_length(self):
        """
        The default length used to automatically create an auxiliary trail edge.

        Returns
        -------
        length : ``float``
            The edge length.
        """
        return self.attributes["_aux_length"]

    @auxiliary_trail_length.setter
    def auxiliary_trail_length(self, length):
        """
        """
        self.attributes["_aux_length"] = length

    @property
    def auxiliary_trail_vector(self):
        """
        The default vector used to automatically create an auxiliary trail edge.

        Returns
        -------
        length : ``float``
            The edge length.
        """
        return self.attributes["_aux_vector"]

    @auxiliary_trail_vector.setter
    def auxiliary_trail_vector(self, vector):
        """
        """
        self.attributes["_aux_vector"] = vector

# ==============================================================================
# Node Additions
# ==============================================================================

    def add_support(self, support):
        """
        Adds a support.

        Parameters
        ----------
        support : ``NodeSupport``
            A node support object.

        Notes
        -----
        Support nodes mark the end of a continuous trail. They aren't fixed.
        """
        value = support.node
        if value is None:
            value = support.xyz
        node = self.node_key(value)

        if node is None:
            raise ValueError("A node doesn't exist at {} yet!".format(value))

        self.node_attribute(node, "type", "support")

    def add_load(self, load):
        """
        Applies a load.

        Parameters
        ----------
        load : ``Load``
            A load object.
        """
        value = load.node
        if value is None:
            value = load.xyz
        node = self.node_key(value)
        if node is None:
            raise ValueError("A node doesn't exist at {} yet!".format(value))

        self.node_attributes(node, ["qx", "qy", "qz"], load.vector)

# ==============================================================================
# Counters
# ==============================================================================

    def number_of_trails(self):
        """
        Number of trails in the topology diagram.

        Return
        ------
        number : ``int``
            The number of trails.
        """
        return len(list(self.trails()))

    def number_of_auxiliary_trails(self):
        """
        Number of auxiliary trails in the topology diagram.

        Return
        ------
        number : ``int``
            The number of auxiliary trails.
        """
        return len(list(self.auxiliary_trails()))

    def number_of_deviation_edges(self):
        """
        Number of deviation edges in the topology diagram.

        Return
        ------
        number : ``int``
            The number of deviation edges.
        """
        return len(list(self.deviation_edges()))

    def number_of_trail_edges(self):
        """
        Number of trail edges in the topology diagram.

        Return
        ------
        number : ``int``
            The number of trail edges.
        """
        return len(list(self.trail_edges()))

# ==============================================================================
# Trails
# ==============================================================================

    def trails(self):
        """
        Iterate over all the existing trails in the topology diagram.

        Yields
        ------
        trail : ``list``
            The next trail.

        Notes
        -----
        A trail is an ordered sequence of node keys.
        This iterator include auxiliary trails, if any.
        """
        return self.attributes["_trails"].values()

    def auxiliary_trails(self):
        """
        Iterate over all the available auxiliary trails in the topology diagram.

        Yields
        ------
        auxiliary_trail : ``list``
            The next auxiliary trail.

        Notes
        -----
        An auxiliary trail is a trail with at least two nodes.
        """
        return self.attributes["_auxiliary_trails"].values()

    def build_trails(self, auxiliary_trails=False):
        """
        Automatically generate the trails in the topology diagram.

        The process starts a graph traversal that begins from the support nodes in the diagram.
        A trail is an ordered sequence of nodes with two characteristics:
        - An origin node is the first node in the trail
        - A support node is the last node in the trail

        Parameters
        ----------
        auxiliary_trails : ``bool``
            A flag to automatically append auxiliary trails to trail-unassigned nodes.

        Notes
        -----
            Origin nodes are computed in automatic as part of the trail-making process.
            Previous trails and auxiliary trails are recalculated every time
            this function is called.
        """
        tr = {}

        # trail search
        nodes_in_trails = set()

        for support in self.support_nodes():

            trail = []
            visited = set()
            node = support

            while True:

                last_node = node
                neighbors = self.neighbors(node)

                while neighbors:
                    neighbor = neighbors.pop()

                    if neighbor in visited:
                        continue

                    try:
                        is_trail = self.is_trail_edge((node, neighbor))
                    except KeyError:
                        is_trail = self.is_trail_edge((neighbor, node))

                    if not is_trail:
                        continue

                    trail.append(node)
                    visited.add(node)
                    node = neighbor
                    break

                if last_node == node:
                    origin = node
                    trail.append(origin)
                    visited.add(node)
                    break

            # set last node to be origin/start node
            self.node_attribute(origin, "type", "_origin")

            trail.reverse()

            # assign node sequences
            # start should be _k= 0, support _k=len(trail)
            for index, node in enumerate(trail):
                self.node_attribute(node, "_k", index)

            tr[origin] = tuple(trail)
            nodes_in_trails.update(visited)

        # output sanity checks
        # all nodes must belong to a trail
        unassigned = set(self.nodes()) - nodes_in_trails

        # automatically create auxiliary trails
        if auxiliary_trails:

            aux_trails = dict()
            aux_dir = normalize_vector(self.auxiliary_trail_vector)

            for node in unassigned:
                aux_vector = scale_vector(aux_dir, self.auxiliary_trail_length)
                aux_xyz = add_vectors(self.node_coordinates(node), aux_vector)
                aux_node = self.add_node(Node(xyz=aux_xyz))

                self.add_support(NodeSupport(aux_node))
                edge = self.add_edge(TrailEdge(node, aux_node, self.auxiliary_trail_length))
                aux_trails[node] = edge

            self.attributes["_auxiliary_trails"] = aux_trails

            print("Warning: {} auxiliary trails have been added to the diagram".format(len(aux_trails)))

            return self.build_trails(auxiliary_trails=False)

        # sanity checks
        # there must be at least one trail edge
        assert len(list(self.trail_edges())) > 0, "No trail edges defined!"
        # there must be at least one support node for trails to run
        assert len(list(self.support_nodes())) > 0, "No supports assigned!"
        # no free nodes
        msg = "Nodes {} haven't been assigned to a trail. Check your topology!".format(unassigned)
        assert len(unassigned) == 0, msg

        # store trails in topology diagram
        self.attributes["_trails"] = tr

# ==============================================================================
#  Node Collections
# ==============================================================================

    def origin_nodes(self):
        """
        Starting nodes of all trails.

        Yields
        -------
        origin_node : ``int``
            The key of the next origin node.
        """
        return self.nodes_where({"type": "_origin"})

# ==============================================================================
#  Connected Edges
# ==============================================================================

    def connected_deviation_edges(self, node):
        """
        Finds the deviation edges connected to a node.

        Parameters
        ----------
        node : ``int``
            A node key.

        Returns
        -------
        deviation_edges : ``list``
            The keys of the connected deviation edges.
            If no deviation edge is attached, the list will be empty.
        """
        return self._connected_edges_predicate(node, self.is_deviation_edge)

    def connected_trail_edges(self, node):
        """
        Finds the trail edges connected to a node.

        Parameters
        ----------
        node : ``int``
            A node key.

        Returns
        -------
        trail_edges : ``list``
            The keys of the connected trail edges.
            If no trail edge is attached, the list will be empty.
        """
        return self._connected_edges_predicate(node, self.is_trail_edge)

    def _connected_direct_deviation_edges(self, node):
        """
        Finds the direct deviation edges connected to a node.

        Parameters
        ----------
        node : ``int``
            A node key.

        Returns
        -------
        deviation_edges : ``list``
            The keys of the connected deviation edges.
            If no deviation edge is attached, the list will be empty.

        Note
        ----
        Direct deviation edges have both end-nodes with equal topological
        distance to a root node. Distances must be precomputed.
        """
        return self._connected_edges_predicate(node, self._is_direct_deviation_edge)

    def _connected_indirect_deviation_edges(self, node):
        """
        Finds the indirect deviation edges connected to a node.

        Parameters
        ----------
        node : ``int``
            A node key.

        Returns
        -------
        deviation_edges : ``list``
            The keys of the connected deviation edges.
            If no deviation edge is attached, the list will be empty.

        Note
        ----
        Indirect deviation edges have both end-nodes with unequal topological
        distance to a root node. Distances must be precomputed.
        """
        return self._connected_edges_predicate(node, self._is_indirect_deviation_edge)

    def _connected_edges_predicate(self, node, predicate):
        """
        Finds the edges connected to a node.

        Parameters
        ----------
        node : ``int``
            A node key.
        predicate : ``func``
            A predicate function to search for a specific edge type.

        Returns
        -------
        selected_edges : ``list``
            The keys of the selected edges.
            If no edge of the given type is attached, the list will be empty.
        """
        deviation_edges = []
        for edge in self.connected_edges(node):
            if predicate(edge):
                deviation_edges.append(edge)
        return deviation_edges

# ==============================================================================
# Edges
# ==============================================================================

    def trail_edges(self, data=False):
        """
        Iterates over the keys of all trail edges.

        Parameters
        ----------
        data : ``bool``
            ``True`` if the edges attributes should be yielded simultaneously.
            Defaults to ``False``.

        Yields
        -------
        trail_edge : ``tuple``
            The key of the next trail edge.
        attributes : ``dict``
            The attributes of the next trail edge if ``data=True``.
        """
        return self.edges_where({"type": "trail"}, data)

    def deviation_edges(self, data=False):
        """
        Iterates over the keys of all deviation edges.

        Parameters
        ----------
        data : ``bool``
            ``True`` if the edges attributes should be yielded simultaneously.
            Defaults to ``False``.

        Yields
        -------
        deviation_edge : ``tuple``
            The key of the next trail edge.
        attributes : ``dict``
            The attributes of the next deviation edge if ``data=True``.
        """
        return self.edges_where({"type": "deviation"}, data)

    def auxiliary_trail_edges(self, data=False):
        """
        Iterates over the keys of all trail edges in the auxiliary trails.

        Parameters
        ----------
        data : ``bool``
            ``True`` if the edges attributes should be yielded simultaneously.
            Defaults to ``False``.

        Yields
        -------
        deviation_edge : ``tuple``
            The key of the next trail edge.
        attributes : ``dict``
            The attributes of the next trail edge if ``data=True``.
        """
        def predicate(edge, attr):
            return self.is_auxiliary_trail_edge(edge)

        return self.edges_where_predicate(predicate, data)

# ==============================================================================
# Node Filters
# ==============================================================================

    def is_node_origin(self, node):
        """
        Checks if a node is an origin node.

        Parameters
        ----------
        node : ``int``
            A node key.

        Returns
        -------
        flag : ``bool``
            ``True``if the node is a origin node.
            ``False`` otherwise.
        """
        return self.node_attribute(key=node, name="type") == "_origin"

# ==============================================================================
# Edge Predicates
# ==============================================================================

    def is_trail_edge(self, edge):
        """
        Tests whether or not an edge is a trail edge.

        Parameters
        ----------
        edge : ``tuple``
            The key of the edge to test.

        Returns
        -------
        flag : ``bool``
            ``True``if the edge is a trail edge. ``False`` otherwise.
        """
        if self.edge_attribute(edge, "type") == "trail":
            return True
        return False

    def is_deviation_edge(self, edge):
        """
        Tests whether or not an edge is a deviation edge.

        Parameters
        ----------
        edge : ``tuple``
            The key of the edge to test.

        Returns
        -------
        flag : ``bool``
            ``True``if the edge is a deviation edge. ``False`` otherwise.
        """
        if self.edge_attribute(edge, "type") == "deviation":
            return True
        return False

    def is_auxiliary_trail_edge(self, edge):
        """
        Tests whether or not an edge is the edge of an auxiliary trail.

        Parameters
        ----------
        edge : ``tuple``
            The key of the edge to test.

        Returns
        -------
        flag : ``bool``
            ``True``if the edge is in an auxiliary trail. ``False`` otherwise.
        """
        if edge in set(self.auxiliary_trails()):
            return True
        return False

    def _is_direct_deviation_edge(self, edge):
        """
        Tests if a deviation edge is direct.

        Parameters
        ----------
        edge : ``tuple``
            The key of the edge to test.

        Returns
        -------
        flag : ``bool``
            ``True``if the deviation edge is direct.
            ``False`` otherwise.
        """
        def predicate(x):
            a, b = self.edge_sequence(edge)
            if a == b:
                return True

        return self._is_deviation_edge_predicate(edge, predicate)

    def _is_indirect_deviation_edge(self, edge):
        """
        Tests if a deviation edge is indirect.

        Parameters
        ----------
        edge : ``tuple``
            The key of the edge to test.

        Returns
        -------
        flag : ``bool``
            ``True``if the deviation edge is indirect.
            ``False`` otherwise.
        """
        def predicate(x):
            a, b = self.edge_sequence(edge)
            if a != b:
                return True

        return self._is_deviation_edge_predicate(edge, predicate)

    def _is_deviation_edge_predicate(self, edge, predicate):
        """
        Tests whether or not a deviation edge fulfills a set of conditions.

        Parameters
        ----------
        edge : ``tuple``
            The key of the edge to test.
        predicate : ``func``
            A function that for user-defined test conditions.
            Predicate must take in a single edge key as argument.

        Returns
        -------
        flag : ``bool``
            ``True``if the deviation edge meets the predicate condition.
            ``False`` otherwise.

        Notes
        -----
        Similar to ``FormDiagram.edges_where_predicate()``.
        """
        if not self.is_deviation_edge(edge):
            return False
        if predicate(edge):
            return True
        return False

# ==============================================================================
#  Sequences
# ==============================================================================

    def node_sequence(self, node):
        """
        Gets the sequence a node is assigned to.

        Parameters
        ----------
        node : ``int``
            The node key.

        Returns
        -------
        k : ``int``
            The sequence key.
        """
        k = self.node_attribute(key=node, name="_k")
        if k is None:
            msg = "Node {} doesn't belong to a sequence yet. Try adding trails first."
            raise ValueError(msg.format(node))

        return k

    def edge_sequence(self, edge):
        """
        Gets the sequence of the nodes of a given edge.

        Parameters
        ----------
        edge : ``tuple``
            The edge key.

        Returns
        -------
        sequences : ``tuple``
            The nodes sequences.
        """
        u, v = edge
        return self.node_sequence(u), self.node_sequence(v)

    def sequences(self):
        """
        Iterate over the sequences in the diagram.

        Yields
        ------
        sequence : `int`
            The next sequence number.
        """
        return range(self.sequence_max())

    def sequence_max(self):
        """
        The largest sequence number.

        Yields
        ------
        sequence : `int`
            The largest sequence number.
        """
        return max([len(trail) for trail in self.trails()])

# ==============================================================================
# Magic methods
# ==============================================================================

    def __repr__(self):
        """
        """
        tpl = "{}(\n\tEdges: {}\n\tTrail edges: {}\n\tDeviation edges: {}\n\tNodes: {}\n\tSupport Nodes: {}\n\tLoaded nodes: {}\n\tTrails: {}\n\tAuxiliary trails: {}\n\t)"
        data = [self.__class__.__name__,
                self.number_of_edges(),
                self.number_of_trail_edges(),
                self.number_of_deviation_edges(),
                self.number_of_nodes(),
                self.number_of_support_nodes(),
                self.number_of_loaded_nodes(),
                self.number_of_trails(),
                self.number_of_auxiliary_trails()]
        return tpl.format(*data)

# ==============================================================================
# Main
# ==============================================================================


if __name__ == "__main__":
    pass



"""
Search a node key in a topology or a form diagram using a point.
"""
from ghpythonlib.componentbase import executingcomponent as component

from compas.utilities import geometric_key
from compas_rhino.geometry import RhinoPoint


class NodeSearchComponent(component):
    def RunScript(self, diagram, point):
        if point and diagram:
            # node = []
            # for pt in point:
            pt = RhinoPoint.from_geometry(point).to_compas()
            gkey = geometric_key(pt, diagram.tol)
            # node.append(diagram.gkey_node[gkey])
            # return node
            return diagram.gkey_node[gkey]



from compas_cem.elements import Node

from compas.utilities import geometric_key


__all__ = ["NodeMixins"]

# ==============================================================================
# Node Mixins
# ==============================================================================


class NodeMixins(object):
    """
    """
    def add_node(self, node):
        """
        Adds double
        """
        key = node.key
        xyz = node.xyz
        x, y, z = xyz

        node = super(NodeMixins, self).add_node(key=key, x=x, y=y, z=z)
        self.gkey_node[self.gkey(xyz)] = node
        return node

    def node_exists(self, value):
        """
        Checks
        """
        if self.node_key(value) is not None:
            return True
        return False

    def node_key(self, value):
        """
        Gets
        """
        if value in self.node:
            key = value
        else:
            key = self.gkey_node.get(self.gkey(value))
        return key

    def update_node_xyz(self, key, xyz):
        """
        Modifies
        """
        gkey = self.gkey(xyz)
        if gkey in self.gkey_node:
            del self.gkey_node[gkey]
        self.add_node(Node(key, xyz))

    def node_xyz(self, key, xyz=None):
        """
        Gets or sets node coordinates.
        """
        if xyz is None:
            return self.node_coordinates(key)
        self.update_node_xyz(key, xyz)

    def gkey(self, xyz):
        """
        Gets
        """
        return geometric_key(xyz, self.tol)

# ==============================================================================
# Main
# ==============================================================================


if __name__ == "__main__":
    from compas_cem.diagrams import TopologyDiagram

    topology = TopologyDiagram()

    node = topology.add_node(Node())
    xyz = [1.0, 0.0, 0.0]
    topology.node_xyz(node, xyz)

    assert topology.node_xyz(node) == xyz



"""
Create a node from a Rhino point.
"""
from ghpythonlib.componentbase import executingcomponent as component

from compas_cem.elements import Node


class NodeComponent(component):
    def RunScript(self, point):
        if point:
            return Node.from_rhino_point(point)



"""
Displays information about the active COMPAS CEM version and environment.
"""
import compas_bootstrapper
from ghpythonlib.componentbase import executingcomponent as component

import compas_cem


class Info(component):
    def RunScript(self):
        self.Message = 'COMPAS CEM v{}'.format(compas_cem.__version__)
        info = 'COMPAS CEM Version: {}\nEnvironment: {}'
        return (info.format(compas_cem.__version__, compas_bootstrapper.ENVIRONMENT_NAME))



import autograd.numpy as np

from functools import partial

from compas_cem import Data

from compas_cem.equilibrium import static_equilibrium

from compas_cem.equilibrium.force_numpy import equilibrium_state_numpy

from compas_cem.optimization import nlopt_solver
from compas_cem.optimization import grad_autograd
from compas_cem.optimization import objective_function_numpy
from compas_cem.optimization import nlopt_status

from nlopt import RoundoffLimited

__all__ = ["Optimizer"]


# ------------------------------------------------------------------------------
# Optimizer
# ------------------------------------------------------------------------------

class Optimizer(Data):
    """
    An object that modifies a form diagram to meet multiple constraints.
    """
    def __init__(self):
        self.parameters = {}
        self.constraints = {}
        self.x_opt = None
        self.penalty = None
        self.evals = None
        self.gradient_norm = None
        self.status = None

# ------------------------------------------------------------------------------
# Counters
# ------------------------------------------------------------------------------

    def number_of_parameters(self):
        """
        The number of optimization parameters.
        """
        return len(self.parameters)

    def number_of_constraints(self):
        """
        The number of constraints added to the optimizer.
        """
        return len(self.constraints)

# ------------------------------------------------------------------------------
# Mappings
# ------------------------------------------------------------------------------

    def index_parameter(self):
        """
        A dictionary that maps indices to parameter keys.
        """
        return {idx: key for idx, key in enumerate(self.parameters.keys())}

    def parameter_index(self):
        """
        A dictionary that maps parameter keys to indices.
        """
        return {key: idx for key, idx in self.index_parameter()}

# ------------------------------------------------------------------------------
# Additions
# ------------------------------------------------------------------------------

    def add_parameter(self, parameter):
        """
        Adds a parameter to the optimization problem.
        """
        key = (parameter.key(), parameter.attr_name())
        self.parameters[key] = parameter

    def add_constraint(self, constraint):
        """
        Adds a goal constraint.
        """
        key = constraint.key()
        self.constraints[key] = constraint

# ------------------------------------------------------------------------------
# Removals
# ------------------------------------------------------------------------------

    def remove_parameter(self, key):
        """
        Removes an optimization parameter.
        """
        if key not in self.parameters:
            raise KeyError("Parameter not found at object key: {}".format(key))
        del self.parameters[key]

    def remove_constraint(self, key):
        """
        Removes a goal constraint from the optimizer.
        """
        if key not in self.constraints:
            raise KeyError("Constraints not found on object key: {}".format(key))
        del self.constraints[key]

# ------------------------------------------------------------------------------
# Objective Function
# ------------------------------------------------------------------------------

    def objective_func(self, topology, grad_func, tmax, eta):
        """
        Test with autodiff.
        """
        obj_func = objective_function_numpy
        func = partial(self._optimize_form, topology=topology, tmax=tmax, eta=eta)
        return partial(obj_func, x_func=func, grad_func=grad_func)

# ------------------------------------------------------------------------------
# Gradient Function
# ------------------------------------------------------------------------------

    def gradient_func(self, topology, tmax, eta):
        """
        """
        x_func = partial(self._optimize_form, topology=topology, tmax=tmax, eta=eta)
        return partial(grad_autograd, grad_func=x_func)

# ------------------------------------------------------------------------------
# Solver
# ------------------------------------------------------------------------------

    def solve_nlopt(self, topology, algorithm, iters, eps=None, tmax=100, eta=1e-6):
        """
        Solve an optimization problem with NLOpt.
        """
        # test for bad stuff before going any further
        self.check_optimization_sanity()

        # compose gradient and objective functions
        topology_b = topology.copy()
        grad_func = self.gradient_func(topology_b, tmax, eta)
        penalty_func = self.objective_func(topology, grad_func, tmax, eta)

        # generate optimization variables
        x = self.optimization_parameters(topology)

        # extract the lower and upper bounds to optimization variables
        bounds_low, bounds_up = self.optimization_bounds(topology)

        # stack keyword arguments
        hyper_parameters = {"f": penalty_func,
                            "algorithm": algorithm,
                            "dims": self.number_of_parameters(),
                            "bounds_low": bounds_low,
                            "bounds_up": bounds_up,
                            "iters": iters,
                            "eps": eps,
                            "ftol": None}

        # assemble optimization solver
        solver = nlopt_solver(**hyper_parameters)

        # solve optimization problem
        x_opt = None
        try:
            x_opt = solver.optimize(x)
            evals = solver.get_numevals()
            print("Optimization ended correctly!")
            print("Number of evaluations incurred: {}".format(evals))
        except RoundoffLimited:
            print("Optimization was halted because roundoff errors limited progress")
            print("Returned results are generally still useful though!")
            x_opt = self.optimization_parameters(topology)

        # fetch last optimum value of loss function
        loss_opt = solver.last_optimum_value()
        evals = solver.get_numevals()
        status = nlopt_status(solver.last_optimize_result())

        # set optimizer attributes
        self.x_opt = x_opt
        self.penalty = loss_opt
        self.evals = evals
        self.status = status

        # set norm of the gradient
        # np.zeros is a dummy array (signature requirement set by nlopt)
        self.gradient = grad_func(x_opt, np.zeros(x_opt.size))
        self.gradient_norm = np.linalg.norm(self.gradient)

        print("Optimization status: {}".format(status))

        # exit like a champion
        return static_equilibrium(topology)

# ------------------------------------------------------------------------------
# Optimization parameters
# ------------------------------------------------------------------------------

    def optimization_parameters(self, form):
        """
        Creates optimization paremeters array.
        Only one entry in the array per constraint.
        Takes care of keeping the ordering.
        """
        # TODO: This can become an (n, 3) array to acount for root nodes variables?
        x = np.zeros(self.number_of_parameters())

        for index, ckey in self.index_parameter().items():
            constraint = self.parameters[ckey]
            x[index] = constraint.start_value(form)

        return x

    def optimization_bounds(self, form):
        """
        Creates optimization bounds array.
        Only one entry in the array per constraint.
        """
        bounds_low = np.zeros(self.number_of_parameters())
        bounds_up = np.zeros(self.number_of_parameters())

        for index, ckey in self.index_parameter().items():
            parameter = self.parameters[ckey]
            bounds_low[index] = parameter.bound_low(form)
            bounds_up[index] = parameter.bound_up(form)

        return bounds_low, bounds_up

# ------------------------------------------------------------------------------
# Updates
# ------------------------------------------------------------------------------

    def _update_topology_origin_nodes(self, x, topology):
        """
        """
        map_xyz_index = {"x": 0, "y": 1, "z": 2}

        for index, ckey in self.index_parameter().items():

            node = self.parameters[ckey].key()

            # TODO: weak check, needs to be handled differently
            if not isinstance(node, int):
                continue

            # TODO: this check should happen upon assembly, not at calculation?
            if not topology.is_node_origin(node):
                msg = "{} is not a root node. Assigned constraint is invalid!"
                raise ValueError(msg.format(node))

            # TODO: refactor to handle xyz more transparently
            xyz = topology.node_xyz(key=node)
            parameter = self.parameters[ckey]
            j = map_xyz_index[parameter.attr_name()]
            xyz[j] = x[index]

            # self.form.node_xyz(key=node, xyz=xyz)  # form.node_xyz(node, y=x[])?
            topology.node_attributes(key=node, names="xyz", values=xyz)

    def _update_topology_edges(self, x, topology):
        """
        """
        for index, ckey in self.index_parameter().items():

            edge = self.parameters[ckey].key()

            # TODO: weak check, needs to be handled differently
            if isinstance(edge, int):
                continue

            if topology.is_trail_edge(edge):
                name = "length"
            elif topology.is_deviation_edge(edge):
                name = "force"

            value = x[index]

            topology.edge_attribute(key=edge, name=name, value=value)

# ------------------------------------------------------------------------------
# Optimization
# ------------------------------------------------------------------------------

    def _calculate_penalty(self, eq_state):
        """
        """
        penalty = 0.0
        for constraint in self.constraints.values():
            penalty += constraint.penalty(eq_state)

        return penalty

    def _optimize_form(self, parameters, topology, tmax, eta):
        """
        """
        self._update_topology_origin_nodes(parameters, topology)
        self._update_topology_edges(parameters, topology)

        eq_state = equilibrium_state_numpy(topology, tmax, eta)

        return self._calculate_penalty(eq_state)

# ------------------------------------------------------------------------------
# Sanity Check
# ------------------------------------------------------------------------------

    def check_optimization_sanity(self):
        """
        Verify the optimization problem is in its sane mind.
        """
        if len(self.parameters) == 0:
            msg = "No parameters defined. Optimization not possible."
            raise ValueError(msg)

        if len(self.constraints) == 0:
            msg = "No constraints defined. Optimization not possible."
            raise ValueError(msg)

# ------------------------------------------------------------------------------
# Magic methods
# ------------------------------------------------------------------------------

    def __repr__(self):
        """
        """
        tpl = "{} with {} parameters and {} constraints. Status: {}"
        return tpl.format(self.__class__.__name__, self.number_of_parameters(), self.number_of_constraints(), self.status)

# ------------------------------------------------------------------------------
# Main
# ------------------------------------------------------------------------------


if __name__ == "__main__":
    pass



"""
Get reaction forces at the support nodes of a form diagram.
"""
from ghpythonlib.componentbase import executingcomponent as component

import rhinoscriptsyntax as rs


class SupportNodeResultsComponent(component):
    def RunScript(self, form, support_node_keys):
        if form:
            support_node_keys = support_node_keys or list(form.support_nodes())
            reaction_forces = [rs.AddPoint(*form.reaction_force(nd)) for nd in support_node_keys]
            return reaction_forces



from compas_cem.optimization.constraints import VectorConstraint


__all__ = ["PointConstraint"]


class PointConstraint(VectorConstraint):
    """
    Pulls the xyz position of a node to a target point.
    """
    def __init__(self, node=None, point=None, weight=1.0):
        super(PointConstraint, self).__init__(node, point, weight)

    def reference(self, data):
        """
        Gets the reference to compare the target against.
        """
        return data["node_xyz"][self.key()]


if __name__ == "__main__":
    pass



from math import fabs

from compas.geometry import add_vectors
from compas.geometry import length_vector
from compas.geometry import scale_vector
from compas.geometry import normalize_vector
from compas.geometry import translate_points

from compas_ghpython import draw_points
from compas_ghpython import draw_lines
from compas_ghpython import draw_polylines

from compas_ghpython.artists import NetworkArtist

from compas_cem import COLORS


__all__ = ["FormArtist", "TopologyArtist"]


class DiagramArtist(NetworkArtist):
    """
    An artist to draw diagrams.

    Parameters
    ----------
    diagram : :class:`compas_cem.diagrams.Diagram`
        A diagram.
    args : list, optional
        Additional arguments for :class:`compas_ghpython.NetworkArtist`
    kwargs : dict, optional
        Extra named arguments for :class:`compas_ghpython.NetworkArtist`
    """
    def __init__(self, diagram, *args, **kwargs):
        super(DiagramArtist, self).__init__(diagram, *args, **kwargs)

        self.edge_state_colors = {-1: COLORS["compression"],
                                  1: COLORS["tension"],
                                  0: COLORS["edge"]}

        self.float_precision = "3f"
        self._diagram = diagram

    @property
    def diagram(self):
        """
        The diagram to draw.

        Returns
        -------
        form : :class:`compas_cem.diagrams.Diagram`
            A form diagram.
        """
        return self._diagram

    @diagram.setter
    def diagram(self, diagram):
        self._diagram = diagram

    def draw_nodes(self, nodes=None):
        """
        Draw a selection of nodes.

        Parameters
        ----------
        nodes: list, optional
            The selection of nodes that should be drawn.
            Default is ``None``, in which case all nodes are drawn.

        Returns
        -------
        nodes : list of :class:`Rhino.Geometry.Point3d`
        """
        nodes = nodes or list(self.diagram.nodes())
        points = []
        for node in nodes:
            points.append({'pos': self.diagram.node_coordinates(node)})

        return draw_points(points)

    def draw_edges(self, edges=None):
        """
        Draw a selection of edges.

        Parameters
        ----------
        edges : list, optional
            A list of edges to draw.
            The default is ``None``, in which case all edges are drawn.
        Returns
        -------
        edges: list of :class:`Rhino.Geometry.Line`
        """
        edges = edges or list(self.diagram.edges())
        lines = []
        for edge in edges:
            start, end = self.diagram.edge_coordinates(*edge)
            lines.append({'start': start, 'end': end})

        return draw_lines(lines)

    def draw_loads(self, nodes=None, scale=1.0, gap=0.0, min_load=1e-3):
        """
        Draw the loads applied to a selection of nodes.

        Parameters
        ----------
        nodes: list, optional
            A list of nodes to draw the loads upon.
            The default is ``None``, in which case all nodes are considered.
        scale : ``float``
            The scale of the loads.
            Defaults to ``1.0``.
        gap : ``float``
            The offset between a node and the incident load.
            Defaults to ``0.0``.
        min_load: ``float``
            The smallest load magnitude to draw.
            Defaults to ``1e-3``

        Returns
        -------
        loads : list of :class:`Rhino.Geometry.Line`
        """
        nodes = nodes or list(self.diagram.nodes())
        attrs = ["qx", "qy", "qz"]
        shift = {node: False for node in nodes}

        return self._draw_forces(nodes, attrs, scale, shift, gap, min_load)

    def _draw_forces(self, nodes, attrs, scale, shift, gap, tol):
        """
        Base method to draws forces (residuals or loads) on nodes as lines.

        Parameters
        ----------
        nodes: ``list``
            The list of node keys where to draw forces.
        attrs : ``list``
            The attribute names of the force vector to draw.
        scale : ``float``
            The forces scale factor.
        shift : ``bool``
            A flat to shift an arrow one length along its own axis.
        gap : ``float``
            The offset between the node and the force.
        tol : ``float``
            The minimum force magnitude to draw.

        Returns
        -------
        list of :class:`Rhino.Geometry.Line`
        """
        arrows = []

        for node in nodes:
            q_vec = self.diagram.node_attributes(node, attrs)
            q_vec_scaled = scale_vector(q_vec, scale)
            q_vec_norm = normalize_vector(q_vec)
            q_len = length_vector(q_vec)

            if q_len < tol:
                continue

            arrow = {}
            start = self.diagram.node_xyz(node)
            end = add_vectors(start, q_vec_scaled)

            # shift
            gap_arrow = gap
            if shift[node]:
                gap_arrow = (gap + length_vector(q_vec_scaled)) * -1

            gap_vector = scale_vector(q_vec_norm, gap_arrow)
            start, end = translate_points([start, end], gap_vector)

            # create gap
            arrow["start"] = start
            arrow["end"] = end

            arrows.append(arrow)

        return draw_lines(arrows)

    def draw_nodes_support(self, nodes=None):
        """
        Draw a selection of the support nodes in the diagram.

        Parameters
        ----------
        nodes: list, optional
            The selection of support nodes that should be drawn.
            Default is ``None``, in which case all support nodes are drawn.

        Returns
        -------
        nodes : list of :class:`Rhino.Geometry.Point3d`
        """
        nodes = self._collection_keys("support_nodes", "is_node_support", nodes)
        return self.draw_nodes(nodes)

    def _collection_keys(self, iterator_func, filter_func, keys=None):
        """
        Safe choose the keys to draw a particular set of topological objects.

        Parameters
        ----------
        iterator_func : str
            The name of the diagram method that iterats over the objects.
        filter_func : str
            The name of the diagram method to check if a key is an object key.
        keys : list, optional
            The selection of the keys of the objects to draw.

        Returns
        -------
        keys : list of keys
        """
        if keys is None:
            iterator = getattr(self.diagram, iterator_func)
            keys = list(iterator())
        else:
            filterer = getattr(self.diagram, filter_func)
            keys = [key for key in keys if filterer(key)]
        return keys


class FormArtist(DiagramArtist):
    """
    Draw a form diagram with flair.

    Parameters
    ----------
    form_diagram : :class:`compas_cem.diagrams.FormDiagram`
        A form diagram.
    args : list, optional
        Additional arguments for :class:`compas_ghpython.NetworkArtist`
    kwargs : dict, optional
        Extra named arguments for :class:`compas_ghpython.NetworkArtist`
    """
    def __init__(self, form_diagram, *args, **kwargs):
        super(FormArtist, self).__init__(form_diagram, *args, **kwargs)

    def draw_reactions(self, nodes=None, scale=1.0, gap=0.0, min_force=1e-3):
        """
        Draw the reaction forces at a selection of support nodes.

        Parameters
        ----------
        nodes: list, optional
            A list of supports nodes to draw the reaction forces on.
            The default is ``None``, in which case all support nodes are taken.
        scale : ``float``
            The scale of the reaction forces.
            Defaults to ``1.0``.
        gap : ``float``
            The offset between a node and the incident reaction force.
            Defaults to ``0.0``.
        min_load: ``float``
            The smallest reaction force magnitude to draw.
            Defaults to ``1e-3``

        Returns
        -------
        reaction : list of :class:`Rhino.Geometry.Line`
        """
        diagram = self.diagram
        nodes = nodes or list(self.diagram.support_nodes())
        attrs = ["rx", "ry", "rz"]

        # TODO: needs a more robust check for arrow orientation
        # what we need is to know whether the arrow needs a full shift.
        shift = {}
        for node in nodes:
            if not diagram.is_node_support(node):
                continue
            s = False
            forces = [diagram.edge_force(e) for e in diagram.connected_edges(node)]
            max_force = max(forces, key=lambda f: fabs(f))
            if max_force < 0.0:
                s = True
            shift[node] = s

        return self._draw_forces(nodes, attrs, scale, shift, gap, min_force)


class TopologyArtist(DiagramArtist):
    """
    Draw a topology diagram with style.

    Parameters
    ----------
    topology_diagram : :class:`compas_cem.diagrams.TopologyDiagram`
        A topology diagram.
    args : list, optional
        Additional arguments for :class:`compas_ghpython.NetworkArtist`
    kwargs : dict, optional
        Extra named arguments for :class:`compas_ghpython.NetworkArtist`
    """
    def __init__(self, topology_diagram, *args, **kwargs):
        super(TopologyArtist, self).__init__(topology_diagram, *args, **kwargs)

    def draw_nodes_origin(self, nodes=None):
        """
        Draw a selection of the origin nodes in the topology diagram.

        Parameters
        ----------
        nodes: list, optional
            The selection of origin nodes that should be drawn.
            Default is ``None``, in which case all origin nodes are drawn.

        Returns
        -------
        rhino_nodes : list of :class:`Rhino.Geometry.Point3d`
        """
        if nodes is None:
            nodes = list(self.diagram.origin_nodes())
        else:
            nodes = [no for no in nodes if self.diagram.is_node_origin(no)]
        if not len(nodes) == 0:
            return self.draw_nodes(nodes)

    def draw_edges_deviation(self, edges=None):
        """
        Draw a selection of deviation edges.

        Parameters
        ----------
        edges : list, optional
            A list of deviation edges to draw.
            The default is ``None``, in which case all deviation edges are drawn.
        Returns
        -------
        rhino_edges : list of :class:`Rhino.Geometry.Line`
        """

        edges = self._collection_keys("deviation_edges", "is_deviation_edge", edges)
        if not len(edges) == 0:
            return self.draw_edges(edges)

    def draw_edges_trail(self, edges=None):
        """
        Draw a selection of trail edges.

        Parameters
        ----------
        edges : list, optional
            A list of trail edges to draw.
            The default is ``None``, in which case all trail edges are drawn.
        Returns
        -------
        rhino_edges : list of :class:`Rhino.Geometry.Line`
        """
        edges = self._collection_keys("trail_edges", "is_trail_edge", edges)
        if not len(edges) == 0:
            return self.draw_edges(edges)

    def draw_trails(self):
        """
        Draw the trails as Rhino polylines.

        Returns
        -------
        rhino_polylines : list of :class:`Rhino.Geometry.Polyline`
        """
        polylines = []
        for trail in self.diagram.trails():
            p = {}
            p["points"] = [self.diagram.node_xyz(node) for node in trail]
            polylines.append(p)

        return draw_polylines(polylines)
